"""
@author: Meizhu Liang @Imperial College, Mark Neil @Imperial College
"""

import numpy as np
from matplotlib import pyplot as plt
from random import seed
import time
import tifffile

try:
    import cupy as cp
    import_cp = True
    print('cupy imported')
except ImportError:
    import_cp = False


# try:
#     import torch
#     pytorch = True
#     print('pytorch imported')
# except ImportError:
#     pytorch = False




class HexSIM_simulator():
    """
    Class to simulate raw data for Hexagonal SIM at regular and right angles.
    """
    def __init__(self, mode, axial, use_cupy, N, pixel_size, magnification, NA, n, wavelength, npoints, zrange,
                 dz, fwhmz):
        """
                N: Points to use in FFT
                """
        self.axial = axial
        self.use_cupy = use_cupy
        self.N = N  # Points to use in FFT
        self.pixel_size = pixel_size  # Camera pixel size
        self.magnification = magnification  # Objective magnification
        dx = self.pixel_size / self.magnification  # Sampling in lateral plane at the sample in um
        self.NA = NA  # Numerical aperture at sample
        self.n = n  # Refractive index at sample
        self.wavelength = wavelength  # Wavelength in um
        self.npoints = npoints  # Number of random points
        seed(1234)  # set random number generator seed
        # eta is the factor by which the illumination grid frequency exceeds the incoherent cutoff, eta = 1 for normal SIM,
        # eta=sqrt(3) / 2 to maximise resolution without zeros in TF
        # For a normal SIM, maximum resolution extension = 1 + eta
        # carrier is 2 * kmax * eta
        self.eta = self.n / self.NA  # right-angle Hex SIM
        # self.eta = n * np.sqrt(3.0) / 2 / NA  # Hex SIM
        self.zrange = zrange  # distance either side of focus to calculate, in microns, could be arbitrary
        self.dz = dz  # step size in axial direction of PSF
        self.fwhmz = fwhmz  # FWHM of light sheet in z
        self.sigmaz = self.fwhmz / 2.355
        self.dxn = self.wavelength / (4 * self.NA)  # 2 * Nyquist frequency in x and y.
        self.Nn = int(np.ceil(self.N * dx / self.dxn / 2) * 2)  # Number of points at Nyquist sampling, even number
        self.dxn = self.N * dx / self.Nn  # correct spacing
        res = self.wavelength / (2 * self.NA)
        oversampling = res / self.dxn  # factor by which pupil plane oversamples the coherent psf data
        dk = oversampling / (self.Nn / 2)  # Pupil plane sampling
        kx, ky = np.meshgrid(np.linspace(-dk * self.Nn / 2, dk * self.Nn / 2 - dk, self.Nn),
                             np.linspace(-dk * self.Nn / 2, dk * self.Nn / 2 - dk, self.Nn))
        self.kr = np.sqrt(kx ** 2 + ky ** 2)  # Raw pupil function, pupil defined over circle of radius 1.
        csum = sum(sum((self.kr < 1)))  # normalise by csum so peak intensity is 1

        alpha = np.arcsin(self.NA / self.n)
        # Nyquist sampling in z, reduce by 10 % to account for gaussian light sheet
        self.dzn = 0.8 * self.wavelength / (2 * n * (1 - np.cos(alpha)))
        Nz = 2 * np.ceil(self.zrange / self.dz)
        self.dz = 2 * self.zrange / Nz
        self.Nzn = int(2 * np.ceil(self.zrange / self.dzn))
        self.dzn = 2 * self.zrange / self.Nzn
        if Nz < self.Nzn:
            self.Nz = self.Nzn
            self.dz = self.dzn
        self.mode = mode
        self.n_PhaseStep = 7  # number of phase steps

    @property
    def use_cupy(self):
        return (self._use_cupy)

    @use_cupy.setter
    def use_cupy(self, bool):
        if bool:
            if import_cp:
                pass
            else:
                raise ModuleNotFoundError
        self._use_cupy = bool


    def point_cloud(self):
        print("Calculating point cloud")

        rad = 10  # radius of sphere of points
        # Multiply the points several timesto get the enough number
        pointsxn = (2 * np.random.rand(self.npoints * 3, 3) - 1) * [rad, rad, rad]

        pointsxnr = np.sum(pointsxn * pointsxn, axis=1)
        points_sphere = pointsxn[pointsxnr < (rad ** 2), :]  # simulate spheres from cubes
        self.points = points_sphere[(range(self.npoints)), :]
        self.points[:, 2] = self.points[:, 2] / 2  # to make the point cloud for OTF a ellipsoid rather than a sphere

        # plt.figure(20)
        ax = plt.axes(projection='3d')
        scatter_plot = ax.scatter(self.points[:, 0], self.points[:, 1], self.points[:, 2],
                               c=(self.points[:, 0] + self.points[:, 1] + self.points[:, 2]), cmap="spring")
        plt.colorbar(scatter_plot)
        # ax.set_xlabel('X')
        # ax.set_ylabel('Y')
        # ax.set_zlabel('Z')
        # plt.title("Point cloud")

    def phase_tilts(self):
        # Generate phase tilts in frequency space

        xyrange = self.Nn / 2 * self.dxn
        dkxy = np.pi / xyrange
        kxy = np.arange(-self.Nn / 2 * dkxy, (self.Nn / 2) * dkxy, dkxy)
        dkz = np.pi / self.zrange
        kz = np.arange(-self.Nzn / 2 * dkz, (self.Nzn / 2) * dkz, dkz)

        if self.use_cupy:
            self.phasetilts = cp.zeros((self.n_PhaseStep, self.Nzn, self.Nn, self.Nn), dtype=np.complex64)
        else:
            self.phasetilts = np.zeros((self.n_PhaseStep, self.Nzn, self.Nn, self.Nn), dtype=np.complex64)

        print("Calculating pointwise phase tilts")

        start_time = time.time()

        for xx in range(self.n_PhaseStep):
            if self.use_cupy:
                pxyz = cp.zeros((self.Nzn, self.Nn, self.Nn), dtype=np.complex64)
            else:
                pxyz = np.zeros((self.Nzn, self.Nn, self.Nn), dtype=np.complex64)
            for i in range(self.npoints):
                x = self.points[i, 0]
                y = self.points[i, 1]
                z = self.points[i, 2] + self.dz / self.n_PhaseStep * (xx - 1)
                ph = self.eta * 4 * np.pi * self.NA / self.wavelength
                p1 = -xx * 2 * np.pi / self.n_PhaseStep
                # p1 = -xx * 2 * np.pi / 7
                p2 = xx * 4 * np.pi / self.n_PhaseStep
                # p2 = xx * 4 * np.pi / 7
                if self.axial:  # axial polarisation normalised to peak intensity of 1
                    ill = 2 / 9 * (3 / 2 + np.cos(ph * y + p1 - p2) + np.cos(ph * (y - np.sqrt(3) * x) / 2 + p1) +
                                   np.cos(ph * (-y - np.sqrt(3) * x) / 2 + p2))
                else:  # in plane polarisation normalised to peak intensity of 1
                    ill = 2 / 9 * (3 - np.cos(ph * (-x - y) / 2 + p1 - p2) - np.cos(ph * (- x) + p1)
                                   - np.cos(ph * (y - x) / 2 + p2))
                    # ill = 2 / 9 * (3 - np.cos(ph * y + p1 - p2) - np.cos(ph * (y - np.sqrt(3) * x) / 2 + p1)
                    # - np.cos(ph * (-y - np.sqrt(3) * x) / 2 + p2))
                if self.use_cupy:
                    px = cp.array(np.exp(1j * np.single(x * kxy)))
                    py = cp.array(np.exp(1j * np.single(y * kxy)))
                    pz = cp.array(np.exp(1j * np.single(z * kz)) * ill)
                    pxy = cp.array(px[:, np.newaxis] * py)
                    for ii in range(len(kz)):
                        pxyz[ii, :, :] = pxy * pz[ii]
                    self.phasetilts[xx, :, :, :] = self.phasetilts[xx, :, :, :] + pxyz
                else:
                    px = np.exp(1j * np.single(x * kxy))
                    py = np.exp(1j * np.single(y * kxy))
                    pz = np.exp(1j * np.single(z * kz)) * ill
                    pxy = px[:, np.newaxis] * py
                    for ii in range(len(kz)):
                        pxyz[ii, :, :] = pxy * pz[ii]
                    self.phasetilts[xx, :, :, :] = self.phasetilts[xx, :, :, :] + pxyz

        elapsed_time = time.time() - start_time
        print(f'Phase tilts calculation:  {elapsed_time:.3f}s')

    def raw_image_stack(self):
        # Calculates point cloud, phase tilts, 3d psf and otf before the image stack

        self.point_cloud()
        self.phase_tilts()
        print("Calculating 3d psf")
        nz = 0
        psf = np.zeros((self.Nzn, self.Nn, self.Nn))
        pupil = self.kr < 1

        for z in np.arange(-self.zrange, self.zrange - self.dzn, self.dzn):
            c = (np.exp(
                1j * (z * self.n * 2 * np.pi / self.wavelength *
                      np.sqrt((1 - (self.kr * pupil) ** 2 * self.NA ** 2 / self.n ** 2))))) * pupil
            psf[nz, :, :] = abs(np.fft.fftshift(np.fft.ifft2(c))) ** 2 * np.exp(-z ** 2 / 2 / self.sigmaz ** 2)
            nz = nz + 1

        # Normalised so power in resampled psf(see later on) is unity in focal plane
        psf = psf * self.Nn ** 2 / np.sum(pupil) * self.Nz / self.Nzn

        print("Calculating 3d otf")
        psf = psf
        otf = np.fft.fftn(psf)
        aotf = abs(np.fft.fftshift(otf))
        m = max(aotf.flatten())
        aotf_z = []
        for x in range(self.Nzn):
            aotf_z.append(np.sum(aotf[x]))

        print("Calculating raw image stack")
        if self.use_cupy:
            img = cp.zeros((self.Nz * self.n_PhaseStep, self.N, self.N), dtype=np.single)
        else:
            img = np.zeros((self.Nz * self.n_PhaseStep, self.N, self.N), dtype=np.single)
        for i in range(self.n_PhaseStep):
            if self.use_cupy:
                ootf = cp.fft.fftshift(otf) * self.phasetilts[i, :, :, :]
                img[cp.arange(i, self.Nz * self.n_PhaseStep, self.n_PhaseStep), :, :] = cp.abs(cp.fft.ifftn(ootf, (self.Nz, self.N, self.N)))
            else:
                ootf = np.fft.fftshift(otf) * self.phasetilts[i, :, :, :]
                img[np.arange(i, self.Nz * self.n_PhaseStep, self.n_PhaseStep), :, :] = np.abs(np.fft.ifftn(ootf, (self.Nz, self.N, self.N)))
        # OK to use abs here as signal should be all positive.
        # Abs is required as the result will be complex as the fourier plane cannot be shifted back to zero when oversampling.
        # But should reduction in sampling be allowed here(Nz < Nzn)?

        # raw image stack
        if self.use_cupy:
            img = cp.asnumpy(img)

        # raw image sum along z axis
        if self.use_cupy:
            img_sum_z = cp.asnumpy(cp.sum(img, axis=0))
        else:
            img_sum_z = np.sum(img, axis=0)

        # raw image sum along x (or y) axis"
        if self.use_cupy:
            img_sum_x = cp.asnumpy(cp.sum(img, axis=1))
        else:
            img_sum_x = np.sum(img, axis=1)

        # Save generated images
        if self.axial:
            stackfilename = "Raw_img_stack90_512_axial.tif"
        else:
            stackfilename = "Raw_img_stack90_512_inplane.tif"
        if self.use_cupy:
            tifffile.imwrite(stackfilename, cp.asnumpy(img))
        else:
            tifffile.imwrite(stackfilename, img)
        print('Raw image stack saved')

        # Return img_sum_z, img_sum_x, psf_xy, otf_yz, otf_xz, otf_sum_xy (otf sum in x-y plane)
        return img, img_sum_z, img_sum_x, psf[int(self.Nzn / 2 + 10), :, :], np.log(aotf[:, int(self.Nn / 2), :].squeeze() + 0.0001), \
               np.log(aotf[:, :, int(self.Nn / 2)].squeeze() + 0.0001)

if __name__ == '__main__':
    t = HexSIM_simulator(None, 0, 1, 512, 5.5, 60, 1.1, 1.33, 0.52, 500, 7.0,
                 0.4, 3.0)
    # t.point_cloud()
    # t.phase_tilts()
    t.raw_image_stack()



"""
The parent class to simulate raw data for SIM. @author: Meizhu Liang @Imperial College
Some calculations are adapted by work by Mark Neil @Imperial College
"""

from random import seed
import numpy as np
import time
import tifffile

try:
    import cupy as cp
    print('cupy imported')
    import_cp = True
except:
    import_cp = False

try:
    import torch
    print('torch imported')
    import_torch = True
    if torch.has_cuda:
        torch_GPU = True
    else:
        torch_GPU = False
except:
    import_torch = False
    torch_GPU = False

class Base_simulator:
    pol = None  # polarisation
    acc = None  # acceleration
    _tdev = None
    N = 512  # Points to use in FFT
    pixel_size = 5.5  # Camera pixel size
    magnification = 60  # Objective magnification
    NA = 1.1  # Numerical aperture at sample
    n = 1.33  # Refractive index at sample
    wavelength = 0.52  # Wavelength in um
    npoints = 500  # Number of random points
    zrange = 7.0  # distance either side of focus to calculate, in microns, could be arbitrary
    dz = 0.4  # step size in axial direction of PSF
    fwhmz = 3.0  # FWHM of light sheet in z
    random_seed = None


    def initialise(self):
        np.random.seed(self.random_seed)
        # self.seed(1234)  # set random number generator seed
        self.eta = self.n / self.NA  # right-angle Hex SIM
        self.sigmaz = self.fwhmz / 2.355
        self.dx = self.pixel_size / self.magnification  # Sampling in lateral plane at the sample in um
        self.dxn = self.wavelength / (4 * self.NA)  # 2 * Nyquist frequency in x and y.
        self.Nn = int(np.ceil(self.N * self.dx / self.dxn / 2) * 2)  # Number of points at Nyquist sampling, even number
        self.dxn = self.N * self.dx / self.Nn  # correct spacing
        self.res = self.wavelength / (2 * self.NA)
        oversampling = self.res / self.dxn  # factor by which pupil plane oversamples the coherent psf data
        self.dk = oversampling / (self.Nn / 2)  # Pupil plane sampling
        self.kx, self.ky = np.meshgrid(np.linspace(-self.dk * self.Nn / 2, self.dk * self.Nn / 2 - self.dk, self.Nn),
                             np.linspace(-self.dk * self.Nn / 2, self.dk * self.Nn / 2 - self.dk, self.Nn))
        self.kr = np.sqrt(self.kx ** 2 + self.ky ** 2)  # Raw pupil function, pupil defined over circle of radius 1.
        self.csum = sum(sum((self.kr < 1)))  # normalise by csum so peak intensity is 1

        self.alpha = np.arcsin(self.NA / self.n)
        # Nyquist sampling in z, reduce by 10 % to account for gaussian light sheet
        self.dzn = 0.8 * self.wavelength / (2 * self.n * (1 - np.cos(self.alpha)))
        self.Nz = int(2 * np.ceil(self.zrange / self.dz))
        self.dz = 2 * self.zrange / self.Nz
        self.Nzn = int(2 * np.ceil(self.zrange / self.dzn))
        self.dzn = 2 * self.zrange / self.Nzn
        if self.Nz < self.Nzn:
            self.Nz = self.Nzn
            self.dz = self.dzn
        self._tdev = torch.device('cuda' if self.acc == 3 else 'cpu')

    def point_cloud(self):

        rad = 10  # radius of sphere of points
        # Multiply the points several times to get the enough number
        pointsxn = (2 * np.random.rand(self.npoints * 3, 3) - 1) * [rad, rad, rad]

        pointsxnr = np.sum(pointsxn * pointsxn, axis=1)  # multiple times the points
        points_sphere = pointsxn[pointsxnr < (rad ** 2), :]  # simulate spheres from cubes
        self.points = points_sphere[(range(self.npoints)), :]
        self.points[:, 2] = self.points[:, 2] / 2  # to make the point cloud for OTF a ellipsoid rather than a sphere

    def phase_tilts(self):
        """Generate phase tilts in frequency space"""
        self._nsteps = self._phaseStep * self._angleStep
        xyrange = self.Nn / 2 * self.dxn
        dkxy = np.pi / xyrange
        dkz = np.pi / self.zrange
        self.kxy = np.arange(-self.Nn / 2 * dkxy, (self.Nn / 2) * dkxy, dkxy)
        self.kz = np.arange(-self.Nzn / 2 * dkz, (self.Nzn / 2) * dkz, dkz)

        if self.acc == 0:
            self.phasetilts = np.zeros((self._nsteps, self.Nzn, self.Nn, self.Nn), dtype=np.complex64)
        elif self.acc == 1:
            self.phasetilts = cp.zeros((self._nsteps, self.Nzn, self.Nn, self.Nn), dtype=np.complex64)
        else:
            self.phasetilts = torch.empty((self._nsteps, self.Nzn, self.Nn, self.Nn), dtype=torch.complex64, device=self._tdev)

        start_time = time.time()

        itcount = 0
        total_its = self._angleStep * self._phaseStep * self.npoints
        lastProg = -1

        for astep in range(self._angleStep):
            for pstep in range(self._phaseStep):
                for i in range(self.npoints):
                    prog = (100 * itcount) // total_its
                    if prog > lastProg:
                        lastProg = prog
                        yield f'Phase tilts calculation: {prog:.1f}% done'
                    itcount += 1
                    isteps = pstep + self._angleStep * astep  # index of the steps
                    self.x = self.points[i, 0]
                    self.y = self.points[i, 1]
                    z = self.points[i, 2] + self.dz / self._nsteps * (isteps)
                    self.ph = self.eta * 4 * np.pi * self.NA / self.wavelength
                    self.p1 = pstep * 2 * np.pi / self._phaseStep
                    self.p2 = -pstep * 4 * np.pi / self._phaseStep
                    self._ill()  # gets illumination from the child class
                    if self.pol == 1:
                        ill = self._illAx
                    elif self.pol == 2:
                        ill = self._illCi
                    else:
                        ill = self._illIp
                    if self.acc == 0:
                        px = np.exp(1j * np.single(self.x * self.kxy))[:, np.newaxis]
                        py = np.exp(1j * np.single(self.y * self.kxy))
                        pz = (np.exp(1j * np.single(z * self.kz)) * ill)[:, np.newaxis, np.newaxis]
                        self.phasetilts[isteps, :, :, :] += (px * py) * pz
                    elif self.acc == 1:
                        px = cp.array(np.exp(1j * np.single(self.x * self.kxy))[:, np.newaxis])
                        py = cp.array(np.exp(1j * np.single(self.y * self.kxy)))
                        pz = cp.array((np.exp(1j * np.single(z * self.kz)) * ill)[:, np.newaxis, np.newaxis])
                        self.phasetilts[isteps, :, :, :] += (px * py) * pz
                    else:
                        px = torch.as_tensor(np.exp(1j * np.single(self.x * self.kxy)), device=self._tdev)
                        py = torch.as_tensor(np.exp(1j * np.single(self.y * self.kxy)), device=self._tdev)
                        pz = torch.as_tensor((np.exp(1j * np.single(z * self.kz)) * ill),
                                          device=self._tdev)
                        self.phasetilts[isteps, :, :, :] += (px[..., None] * py) * pz[..., None, None]

        self.elapsed_time = time.time() - start_time
        yield f'Phase tilts calculation time:  {self.elapsed_time:3f}s'

    def raw_image_stack(self):
        # Calculates point cloud, phase tilts, 3d psf and otf before the image stack
        self.initialise()
        self.point_cloud()
        yield "Point cloud calculated"

        for msg in self.phase_tilts():
            yield msg

        # Calculating psf
        nz = 0
        psf = np.zeros((self.Nzn, self.Nn, self.Nn))
        pupil = self.kr < 1
        for z in np.arange(-self.zrange, self.zrange - self.dzn, self.dzn):
            c = (np.exp(
                1j * (z * self.n * 2 * np.pi / self.wavelength *
                      np.sqrt(1 - (self.kr * pupil) ** 2 * self.NA ** 2 / self.n ** 2)))) * pupil
            psf[nz, :, :] = abs(np.fft.fftshift(np.fft.ifft2(c))) ** 2 * np.exp(-z ** 2 / 2 / self.sigmaz ** 2)
            nz = nz + 1
        # Normalised so power in resampled psf(see later on) is unity in focal plane
        psf = psf * self.Nn ** 2 / np.sum(pupil) * self.Nz / self.Nzn
        self.psf_z0 = psf[int(self.Nzn / 2 + 10), :, :]  # psf at z=0
        yield "psf calculated"

        # Calculating 3d otf
        otf = np.fft.fftn(psf)
        aotf = abs(np.fft.fftshift(otf))  # absolute otf
        m = max(aotf.flatten())
        aotf_z = []
        for x in range(self.Nzn):
            aotf_z.append(np.sum(aotf[x]))
        self.aotf_x = np.log(
            aotf[:, int(self.Nn / 2), :].squeeze() + 0.0001)  # cross section perpendicular to x axis
        self.aotf_y = np.log(aotf[:, :, int(self.Nn / 2)].squeeze() + 0.0001)
        yield "3d otf calculated"

        if self.acc == 0:
            img = np.zeros((self.Nz * self._nsteps, self.N, self.N), dtype=np.single)
        elif self.acc == 1:
            img = cp.zeros((self.Nz * self._nsteps, self.N, self.N), dtype=np.single)
        else:
            img =torch.empty((self.Nz * self._nsteps, self.N, self.N), dtype=torch.float, device=self._tdev)

        for i in range(self._nsteps):
            if self.acc == 0:
                ootf = np.fft.fftshift(otf) * self.phasetilts[i, :, :, :]
                img[np.arange(i, self.Nz * self._nsteps, self._nsteps), :, :] = abs(
                    np.fft.ifftn(ootf, (self.Nz, self.N, self.N)))
            elif self.acc == 1:
                ootf = cp.fft.fftshift(otf) * self.phasetilts[i, :, :, :]
                img[cp.arange(i, self.Nz * self._nsteps, self._nsteps), :, :] = cp.abs(
                    cp.fft.ifftn(ootf, (self.Nz, self.N, self.N)))
            else:
                ootf = torch.fft.fftshift(torch.as_tensor(otf, device=self._tdev), ) * self.phasetilts[i, :, :, :]
                img[torch.arange(i, self.Nz * self._nsteps, self._nsteps), :, :] = (torch.abs(
                    torch.fft.ifftn(ootf, (self.Nz, self.N, self.N)))).to(torch.float)
        # OK to use abs here as signal should be all positive.
        # Abs is required as the result will be complex as the fourier plane cannot be shifted back to zero when oversampling.
        # But should reduction in sampling be allowed here(Nz < Nzn)?

        stackfilename = f"Raw_img_stack_{self.N}_{self.pol}.tif"
        if self.acc == 0:
            # raw image stack
            # raw image sum along z axis
            self.img_sum_z = np.sum(img, axis=0)
            # raw image sum along x (or y) axis
            self.img_sum_x = np.sum(img, axis=1)
            self.img = img
        elif self.acc == 1:
            self.img_sum_z = cp.asnumpy(cp.sum(img, axis=0))
            self.img_sum_x = cp.asnumpy(cp.sum(img, axis=1))
            self.img = cp.asnumpy(img)
        elif self.acc == 2:
            self.img_sum_z = (torch.sum(img, axis=0)).numpy()
            self.img_sum_x = (torch.sum(img, axis=1)).numpy()
            self.img = img.numpy()
        elif self.acc == 3:
            self.img_sum_z = (torch.sum(img, axis=0)).detach().cpu().numpy()
            self.img_sum_x = (torch.sum(img, axis=1)).detach().cpu().numpy()
            self.img = img.detach().cpu().numpy()

        # Save generated images
        tifffile.imwrite(stackfilename, self.img)
        print('Raw image stack saved')

        yield f'Finished, Phase tilts calculation time:  {self.elapsed_time:3f}s'
        print(f'Finished, Phase tilts calculation time:  {self.elapsed_time:3f}s')



"""
@authors: Meizhu Liang @Imperial College
"""
from magicgui import magicgui
from magicgui.widgets import Container
from enum import Enum
from napari_generic_simulator.baseSIMulator import import_cp, import_torch, torch_GPU
from napari_generic_simulator.hexSIMulator import HexSim_simulator, RightHexSim_simulator
from napari_generic_simulator.conSIMulator import ConSim_simulator
from qtpy.QtWidgets import QWidget, QVBoxLayout, QLineEdit
from napari.qt.threading import thread_worker

class Sim_mode(Enum):
    HEXSIM = 0
    HEXSIM_RIGHT_ANGLES = 1
    SIM_CONV = 2  # Conventional 2-beam SIM

class Pol(Enum):
    IN_PLANE = 0
    AXIAL = 1
    CIRCULAR = 2

class Accel(Enum):
    USE_NUMPY = 0
    if import_cp:
        USE_CUPY = 1
    if import_torch:
        USE_TORCH_CPU = 2
        if torch_GPU:
            USE_TORCH_GPU = 3

class SIMulator(QWidget):
    """
    A Napari plugin for the simulation of raw images produced while scanning an object (3D point cloud) through focus as
    the hexSIM illumination pattern is shifted through 7 positions laterally.The raw data is processed by a standard, a
    frame-by-frame and a batch reconstruction to produce the super-resolved output.https://doi.org/10.1098/rsta.2020.0162
    This currently supports hexagonal SIM (1 angle, 7 phases) with three beams and that at right-angles.
    """
    def __init__(self, viewer: 'napari.viewer.Viewer'):
        self._viewer = viewer
        super().__init__()
        self.parameters()
        self.setup_ui()
        self.start_simulator()

    def setup_ui(self):
        self.wrap_widgets()
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.add_magic_function(self.w, layout)
        self.messageBox = QLineEdit()
        layout.addWidget(self.messageBox, stretch=True)
        self.messageBox.setText('Messages')

    def add_magic_function(self, function, _layout):
        self._viewer.layers.events.inserted.connect(function.reset_choices)
        self._viewer.layers.events.removed.connect(function.reset_choices)
        _layout.addWidget(function.native)

    def parameters(self, SIM_mode=Sim_mode.HEXSIM_RIGHT_ANGLES, Polarisation=Pol.AXIAL, Acceleration=list(Accel)[-1],
                   N: int = 512, pixel_size_μm: float = 5.5, magnification: int = 60, NA: float = 1.1, n: float = 1.33,
                   wavelength_μm: float = 0.52, npoints: int = 500, zrange_μm: float = 7.0, dz_μm: float = 0.35,
                   fwhmz_μm: float = 3.0, random_seed: int = 123):
        self.SIM_mode = SIM_mode.value
        self.Polarisation = Polarisation.value
        self.Acceleration = Acceleration.value
        self.N = N
        self.pixel_size = pixel_size_μm
        self.magnification = magnification
        self.NA = NA
        self.n = n
        self.wavelength = wavelength_μm
        self.npoints = npoints
        self.zrange = zrange_μm
        self.dz = dz_μm
        self.fwhmz = fwhmz_μm
        self.random_seed = random_seed
        self.par_list =[self.SIM_mode, self.Polarisation, self.Acceleration, self.N, self.pixel_size, self.magnification, self.NA, self.n, self.wavelength,
                self.npoints, self.zrange, self.dz, self.fwhmz, self.random_seed]

    def set_att(self):
        """Sets attributes in the simulation class. Executed frequently to update the parameters"""
        if self.SIM_mode == Sim_mode.HEXSIM.value:
            self.sim = HexSim_simulator()
        elif self.SIM_mode ==Sim_mode.HEXSIM_RIGHT_ANGLES.value:
            self.sim = RightHexSim_simulator()
        elif self.SIM_mode == Sim_mode.SIM_CONV.value:
            self.sim = ConSim_simulator()

        if self.Polarisation == Pol.IN_PLANE.value:
            self.sim.pol = 0
        elif self.Polarisation == Pol.AXIAL.value:
            self.sim.pol = 1
        elif self.Polarisation == Pol.CIRCULAR.value:
            self.sim.circular = 2

        if self.Acceleration == Accel.USE_NUMPY.value:
            self.sim.acc = 0
        if hasattr(Accel, 'USE_CUPY'):
            if self.Acceleration == Accel.USE_CUPY.value:
                self.sim.acc = 1
        if hasattr(Accel, 'USE_TORCH_CPU'):
            if self.Acceleration == Accel.USE_TORCH_CPU.value:
                self.sim.acc = 2
        if hasattr(Accel, 'USE_TORCH_GPU'):
            if self.Acceleration == Accel.USE_TORCH_GPU.value:
                self.sim.acc = 3

        self.sim.N = self.N
        self.sim.pixel_size = self.pixel_size
        self.sim.magnification = self.magnification
        self.sim.NA = self.NA
        self.sim.n = self.n
        self.sim.wavelength = self.wavelength
        self.sim.npoints = self.npoints
        self.sim.zrange = self.zrange
        self.sim.dz = self.dz
        self.sim.fwhmz = self.fwhmz
        self.sim.random_seed = self.random_seed

    def start_simulator(self):
        """
        Starts the raw images generators and create the frequency space.

        """
        if hasattr(self, 'sim'):
            self.stop_simulator()
            self.start_simulator()
        else:
            self.set_att()

    def setReconstructor(self):
        pass

    def stop_simulator(self):
        if hasattr(self, 'sim'):
            delattr(self, 'sim')

    def show_img(self):
        self._viewer.add_image(data=self.sim.img, name='raw image stack')

    @thread_worker(connect={"returned": show_img})
    def get_results(self):
        self.set_att()
        t = self.sim.raw_image_stack()
        try:
            while True:
                self.messageBox.setText(next(t))
        except Exception as e:
            print(e)
        self.used_par_list = [self.SIM_mode, self.Polarisation, self.Acceleration, self.N, self.pixel_size,
                              self.magnification, self.NA, self.n, self.wavelength, self.npoints, self.zrange, self.dz,
                              self.fwhmz, self.random_seed]
        return self

    def show_raw_img_sum(self, show_raw_img_sum: bool=False):
        if show_raw_img_sum:
            if hasattr(self.sim, 'img_sum_z'):
                if self.used_par_list != self.par_list:
                    self.messageBox.setText('Parameters changed! Calculate the raw-image stack first!')
                else:
                    try:
                        self._viewer.add_image(self.sim.img_sum_z, name='raw image sum along z axis')
                        self._viewer.add_image(self.sim.img_sum_x, name='raw image sum along x (or y) axis')
                    except Exception as e:
                        print(str(e))

    def show_psf(self, show_3D_psf: bool=False):
        if show_3D_psf:
            if hasattr(self.sim, 'psf_z0'):
                if self.used_par_list != self.par_list:
                    self.messageBox.setText('Parameters changed!')
                else:
                    try:
                        self._viewer.add_image(self.sim.psf_z0, name='PSF in x-y plane')
                    except Exception as e:
                        print(e)

    def show_otf(self, show_3D_otf: bool=False):
        if show_3D_otf:
            if hasattr(self.sim, 'aotf_x'):
                if self.used_par_list != self.par_list:
                    self.messageBox.setText('Parameters changed! Calculate the raw-image stack first!')
                else:
                    try:
                        self._viewer.add_image(self.sim.aotf_x, name='OTF perpendicular to x')
                        self._viewer.add_image(self.sim.aotf_y, name='OTF perpendicular to y')
                    except Exception as e:
                        print(str(e))

    def wrap_widgets(self):
        w1 = magicgui(self.parameters, layout="vertical", auto_call=True)
        w2 = magicgui(self.get_results, call_button="Calculate raw image stack", auto_call=False)
        w3 = magicgui(self.show_raw_img_sum, auto_call=True)
        w4 = magicgui(self.show_psf, layout="vertical", auto_call=True)
        w5 = magicgui(self.show_otf, layout="vertical", auto_call=True)
        self.w = Container(widgets=[w1,w2, w3, w4, w5], labels=None)

if __name__ == '__main__':
    import napari
    viewer = napari.Viewer()
    test = SIMulator(viewer)
    viewer.window.add_dock_widget(test, name='my second app', add_vertical_stretch=True)

    napari.run()